/*
 * Simple reversing trick. Upon correct password entry the flag is printed
 *
 * gcc -O0 -o -s -o littletrick littletrick.c
 *
 * Hint 1: use file and strings tools to perform initial examination of the file
 *
 * Hint 2: there was a hint inside the executable, that strings is not enough. Try
 * to use hexdump or xxd and look closely at the texts, there can be hidden symbols
 *
 * Hint 3: you can enter control characters to the password prompt. the strings
 * in C program are null-terminated. Data is word-aligned (32 or 64 bits), This
 * means, that only one non-printable character between two strings cannot be a
 * separate variable.
 *
 * Flag: UEblAGfl
 */

#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdint.h>
#include "flag.h"

static struct termios orig_termios;
static int ttyfd = STDIN_FILENO;
void tty_reset(void);
void decrypt (uint32_t [2], uint32_t [4]);

int main(int argc, char **argv)
{

	/* The correct password between two other texts to be discoverable w/ xxd */
	/* DO NOT use compiler optimization, to avoid removing the unused hint */

	char *sorry = "Wrong password, no flag. Sorry";
	char *passwd = P;
	char *hint = "strings is not enough to solve the challenge...";

	/* Encrypted flag and decryption key */
	uint32_t v[2] = { V1, V2};
	uint32_t k[4] = { 0x57438952, 0xabde38cb, 0x3243ba30, 0x5635649f };

	/* buf is used to store the answer, keep it long enough to fit the passwd */
	char c, buf[16];
	struct termios raw_termios;
	int i, j;

	/* Set TTY to raw mode to immediately get a character without waiting
	 * for ENTER. To do that, we test if STDIN is actually a terminal,
	 * then save old settings for later and then set the raw mode */

	if (!isatty(ttyfd)) {
		fprintf (stderr, "please run from terminal\n");
		exit (1);
	}

	if (tcgetattr(ttyfd, &orig_termios) < 0) {
		perror ("tcgetattr");
		exit (1);
	}

	if (atexit(tty_reset) < 0) {
		perror ("atexit");
		exit (1);
	}


	raw_termios = orig_termios;
	raw_termios.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
	raw_termios.c_oflag &= ~OPOST;
	raw_termios.c_cflag |= CS8;
	raw_termios.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
	raw_termios.c_cc[VMIN] = 1;
	raw_termios.c_cc[VTIME] = 0;
	if (tcsetattr(ttyfd, TCSAFLUSH, &raw_termios) < 0) {
		perror ("tcsetattr");
		exit (1);
	}

	/* clear the input buffer */
	memset (buf, 0, 16);

	/* now read the password. read is finished at enter (unless enter is
	 * first keypress) or when 16 bytes are read
	 * fflush() is used frequently to force text appear to screen immediately */

	printf ("Password? ");
	fflush (stdout);
	for (i=0; i<16; i++) {
		if (read (ttyfd, &c, 1) < 0) {
			perror ("read");
			tty_reset();
			exit(1);
		}

		/* process bacspace or delete */
		if (c == 0x08 || c == 0x7f) {
			if (i) {
				i-=2;
				printf ("\b \b");
				fflush (stdout);
			}
			else
				i--;
			continue;
		}

		buf[i] = c;

		/* print printable characters and line feed, otherwise print a dot */
		/* break at enter unless it's first character in buf */
		if (c == 0x0d || c == 0x0a) {
			if (i)
				break;
			else
				puts ("\r");	/* \n will be added automatically */
		}
		else if (isprint (c))
			putchar(c);
		else
			putchar('.');
		fflush(stdout);
	}

	/* reset terminal to normal mode and move to next line */
	tty_reset();
	putchar('\n');

	/* compare the results and decrypt flag if successful */
	if (memcmp(passwd, buf, 16) == 0) {
		decrypt (v, k);
		for (i=0; i<4; i++)
			printf("%c%c", (v[0]>>(8*i))&0xff, (v[1]>>(8*i))&0xff);
	}
	else
		printf ("%s", sorry);
	putchar('\n');
	exit(0);
}


/* this is separate function, so it can be used with atexit() */

void tty_reset(void)
{
	(void)tcsetattr(ttyfd, TCSAFLUSH, &orig_termios);
}


/* decrypt flag: v - encrypted text, k - decryption key */

void decrypt (uint32_t v[2], uint32_t k[4]) {
	uint32_t v0=v[0], v1=v[1], sum=0xC6EF3720, i;  /* set up; sum is 32*delta */
	uint32_t delta=0x9E3779B9;                     /* a key schedule constant */
	uint32_t k0=k[0], k1=k[1], k2=k[2], k3=k[3];   /* cache key */

	for (i=0; i<32; i++) {                         /* basic cycle start */
		v1 -= ((v0<<4) + k2) ^ (v0 + sum) ^ ((v0>>5) + k3);
		v0 -= ((v1<<4) + k0) ^ (v1 + sum) ^ ((v1>>5) + k1);
		sum -= delta;
	}                                              /* end cycle */
	v[0]=v0; v[1]=v1;
}

